#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <math.h>

#include "agent.h"
#include "node.h"
#include "Tcl.h"
#include "packet.h"
#include "link.h"
#include "scheduler.h"

#define MAXCLASS	256

extern "C" { int strcasecmp(const char*, const char*); }

class CBRed;

class RClass;

typedef RClass* (RClass::*LinkageMethod)();

class RClass : public NsObject {
public:
	RClass();
  //	virtual int command(int argc, const char*const* argv);
	double	last_update_;
	double	target_bw_;	/* fraction of link bandwidth */
	double	wbwa_;	/* bytes per second allowed this class */
	double	weight_;		/* to plot for debugging */
	double	probability_;
	double	mark_cnt_;
	int	psize_;
	int	satisfied();
	int	markpacket();
	void	log_class_send(Packet* pkt);
};

static class RClassMatcher : public Matcher {
public:
	RClassMatcher() : Matcher("rclass") {}
	TclObject* match(const char* id) {
		return (new RClass);
	}
} matcher_rclass;

RClass::RClass()
{
	/* Does this line belong anywhere?  
	Tcl& tcl = Tcl::instance();
	*/
	link_real("ns_rclass", "weight", &weight_, 0);
	link_bw("ns_rclass", "target-bw", &target_bw_, 0);

	last_update_ = 0;
	mark_cnt_ = 1;
	wbwa_ = 0.99*target_bw_;
	psize_ = 1000;
	probability_ = 1;
}

int RClass::satisfied()
{
	double	now = Scheduler::instance().clock();

	if (wbwa_ < target_bw_) return(0);
	else return(1);
}

void RClass::log_class_send(Packet* pkt)
{
	double	now = Scheduler::instance().clock();
	double bw;
	bw = (pkt->size * 8)/(now-last_update_);
	wbwa_ = weight_*bw + (1-weight_)*wbwa_;
	last_update_ = now;
}

int RClass::markpacket()
{
	mark_cnt_ += probability_;

	if (mark_cnt_ > 1) {
		mark_cnt_--;
		return(1);
	}
	else {
		return(0);
	}
}

void RClass:update_probability()
{
  double cwnd,pwnd;
  double now = Scheduler::instance().clock();
  cwnd = (wbwa_ * 0.5)/(8*psize_);
  pwnd = probability_ * cwnd;
  if (satisfied())
    pwnd -= 1/cwnd;
  else 
    pwnd += 1/cwnd;
  probability_ = pwnd/cwnd;
  if (probability_ > 1) probability_ = 1;
  if (probability_ < 0) probability_ = 0;
}

class CBREDPQueue : public REDPQueue {
public:	
	CBREDPQueue();
	virtual int command(int argc, const char*const* argv);
	Packet	*deque(void);
	void	enque(Packet *pkt);
protected:
	virtual void insert_class(Class*);
	void	plot(Class *cl, double idle);

	Class	*classtab_[MAXCLASS];
	Class	*list_;
};

static class CBREDPQueueMatcher : public Matcher {
public:
	CBREDPQueueMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "cbredp") == 0)
			return (new CBREDPQueue);
		return (0);
	}
} matcher_cbred;

CBREDPQueue::CBREDPQueue()
	: borrowed_(0), list_(0)
{
  //	link_int("ns_cbq", "algorithm", &algorithm_, 0);
  //	link_int("ns_cbq", "max-pktsize", &maxpkt_, 1024);

	int i;
	for (i = 0; i < MAXCLASS; ++i)
		classtab_[i] = 0;
}

/*
 * This implements the following tcl commands:
 *  $cbq insert $class
 *  $cbq bind $class $classID
 */
int CBREDPQueue::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "insert") == 0) {
			Class* p = (Class*)TclObject::lookup(argv[2]);
			insert_class(p);
			return (TCL_OK);
		}
	} else if (argc == 4) {
		if (strcmp(argv[1], "bind") == 0) {
			Class* p = (Class*)TclObject::lookup(argv[2]);
			int id = atoi(argv[3]);
			/*XXX check id range */
			classtab_[id] = p;
			return (TCL_OK);
		}
	}
	return (Link::command(argc, argv));
}

/*
 * Add a class to this cbq object.
 */
void CBREDPQueue::insert_class(Class* p)
{
	p->next_ = list_;
	list_ = p;
}


/* Plot needs to be implemented still. */ 
void CBREDPQueue::plot(Class* cl, double idle)
{
	double t = Scheduler::instance().clock();
   	sprintf(trace_->buffer(), "%g\n", t);
	trace_->dump();
}

/*
 * Map arriving packets to the appropriate class, and add the 
 * arriving packet to the queue for that class.
 * If algorithm==1, and that class could send a packet borrowing 
 * from depth at most d, then "stp->depth" is set to no more than d.
 */
void CBREDPQueue::enque(Packet* pkt)
{
	double	now = Scheduler::instance().clock();
	Class	*cl = classtab_[pkt->class_];
	log_packet_arrival(pkt);

	//
	// note: the following will *SILENTLY DROP* packets which don't
	// belong to a known CBQ class.  this should probably be changed
	// when simulations with unknown traffic are used
	//

	if (cl != 0) {
	  cl->log_class_send(pkt);
	  if (cl->markpacket())
	    pkt->flags |= PF_PRI;
	}

	REDPQueue::enque(pkt);
	return;
}

Packet* CBREDPQueue::deque(void)
{
	double now = Scheduler::instance().clock();
	Packet* pkt = REDPQueue::deque();
	Class	*cl = classtab_[pkt->class_];	
	if ((cl != 0) && (pkt != 0)) {
	  cl->update_probability();
	}
	log_packet_departure(pkt);
	return (pkt);
}
