/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: /n/olympus/opt/users/CVS/ns-1.1-master/tcp-newreno.cc,v 1.5 1997/09/24 15:55:17 wuchang Exp $ (LBL)";
#endif

//
// newreno-tcp: a revised reno TCP source, without sack
//

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"
#include "trace.h"
#include "random.h"
#include "node.h"
#include "scoreboard.h"

class NewRenoTcpAgent : public TcpAgent {
 public:
	NewRenoTcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
 protected:
	u_int dupwnd_;
	int changes_;		/* 0 for fixing unnecessary fast retransmits */
				/* 1 for additional code from Allman, */
				/*  to implement other algorithms from */
				/*  Hoe's paper */
	int ecn_fix_;		/* 0 for buggy opencwnd after quench */
				/* 1 for non-buggy opencwnd after quench */
	void partialnewack(Packet *pkt);
};

static class NewRenoTcpMatcher : public Matcher {
public:
	NewRenoTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno") == 0)
			return (new NewRenoTcpAgent());
		return (0);
	}
} matcher_reno;

int NewRenoTcpAgent::window()
{
        //
        // reno: inflate the window by dupwnd_
        //      dupwnd_ will be non-zero during fast recovery,
        //      at which time it contains the number of dup acks
        //
	int win = int(cwnd_) + dupwnd_;
        if (win > int(wnd_))
                win = int(wnd_);
        return (win);
}

NewRenoTcpAgent::NewRenoTcpAgent() : dupwnd_(0), changes_(0)
{
	link_int("ns_tcpnewreno", "changes", &changes_, 0);
	link_int("ns_tcpnewreno", "ecn-fix", &ecn_fix_, 0);
}

/* 
 * Process a packet that acks previously unacknowleges data, but
 * does not take us out of Fast Retransmit.
 */
void NewRenoTcpAgent::partialnewack(Packet* pkt)
{
        newtimer(pkt);
#ifdef notyet
        if (pkt->seqno_ == stp->maxpkts && stp->maxpkts > 0)
                stp->endtime = (float) realtime();
#endif
        last_ack_ = pkt->seqno_;
        highest_ack_ = last_ack_;
        if (seqno_ < last_ack_ + 1)
                seqno_ = last_ack_ + 1;
        if (rtt_active_ && pkt->seqno_ >= rtt_seq_) {
                rtt_active_ = 0;
                t_backoff_ = 1;
        }
}

void NewRenoTcpAgent::recv(Packet *pkt)
{
	static int acked = 0;
	int new_ssthresh;
	static double ack2, ack3, basertt;

        /* Use first packet to calculate the RTT  --contributed by Allman */

        if (++acked == 1)
                basertt = Scheduler::instance().clock() - firstsent_;

        /* Estimate ssthresh based on the calculated RTT and the estimated
           bandwidth (using ACKs 2 and 3).  */

        else if (acked == 2)
                ack2 = Scheduler::instance().clock();
        else if (acked == 3)
        {       
                ack3 = Scheduler::instance().clock();  
                new_ssthresh = int((basertt * (size_ / (ack3 - ack2)))
                                    / size_);
                if (changes_ > 0 && new_ssthresh < ssthresh_)
                    ssthresh_ = new_ssthresh;
        }

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
	if (pkt->seqno_ > last_ack_) {
	    if (pkt->seqno_ >= recover_) {
		dupwnd_ = 0;
		newack(pkt);
		if (!(pkt->flags_ & PF_ECN) || !ecn_fix_)
                	opencwnd();
	    }
	    else {
		/* received new ack for a packet sent during Fast
		 *  Recovery, but sender stays in Fast Recovery */
		dupwnd_ = 0;
		partialnewack(pkt);
		output(last_ack_ + 1, 0);
	    }
   	} else if (pkt->seqno_ == last_ack_)  {
		if (++dupacks_ == NUMDUPACKS) {
			/*
			 * Assume we dropped just one packet.
			 * Retransmit last ack + 1
			 * and try to resume the sequence.
			 */
                       /* The line below, for "bug_fix_" true, avoids
                        * problems with multiple fast retransmits after
			* a retransmit timeout.
                        */
			if ( (highest_ack_ > recover_) ||
			    ( recover_cause_ != 2)) {
				recover_cause_ = 1;
				recover_ = maxseq_;
				closecwnd(1);
				reset_rtx_timer(1);
				output(last_ack_ + 1, PF_DUPACK);
                        }
			dupwnd_ = NUMDUPACKS;
		} else if (dupacks_ > NUMDUPACKS) 
			++dupwnd_;
                        /* For every two duplicate ACKs we receive (in the
                         * "fast retransmit phase"), send one entirely new
                         * data packet "to keep the flywheel going".  --Allman 
			 */
			 if (changes_ > 0 && (dupacks_ % 2) == 1)
		   		output (seqno_++,0);

	}
	Packet::free(pkt);
	if (trace_)
		plot();

	/*
	 * Try to send more data
	 */

	if (dupacks_ == 0) 
                /*
                 * Maxburst is really only needed for the first
                 *   window of data on exiting Fast Recovery.
                 */
		send(0, 0, maxburst_);
	else if (dupacks_ > NUMDUPACKS - 1 && changes_==0)
		send(0, 0, 2);
}

void NewRenoTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
	}
	TcpAgent::timeout(tno);
}

// NewRenoPolice, NewRenoCload, NewRenoTimed, NewRenoTimedCload by
// Wu-chang Feng <wuchang@eecs.umich.edu>

class NewRenoPoliceTcpAgent : public NewRenoTcpAgent {
public:
	NewRenoPoliceTcpAgent();
	virtual void output(int seqno, int reason);
	virtual void reset();
	void send(int force, int reason, int maxburst = 0);
	void update_bucket();
protected:
	double lbrate_;         /* Leaky bucket rate (bits/sec) */
	double lbdepth_;        /* Leaky bucket depth (seconds) */
	double timer_interval_; /* Fine-grained timer-interval */
	int compliant_;
	int non_compliant_;
	double lbdepth_bytes_;	 // Leaky bucket depth (bytes)
	double last_arriv_;      // Last time a packet arrived
	double tokens_avail_;    // Tokens left (bytes)
	double tokens_wasted_;
	void plot();
};

static class NewRenoPoliceTcpMatcher : public Matcher {
public:
	NewRenoPoliceTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno-police") == 0)
			return (new NewRenoPoliceTcpAgent());
		return (0);
	}
} matcher_renopolice;

NewRenoPoliceTcpAgent::NewRenoPoliceTcpAgent()
{
	link_bw("ns_tcp_newreno_police", "lbrate", &lbrate_, 0);
	link_time("ns_tcp_newreno_police","lbdepth",&lbdepth_, 0);
	link_time("ns_tcp_newreno_police","timer-interval",&timer_interval_, 0);

	link_real(0, "lbdepth-bytes", &lbdepth_bytes_, 1);
	link_real(0, "last-arrival", &last_arriv_, 1);
	link_real(0, "tokens", &tokens_avail_, 1);
	link_real(0, "tokens-wasted", &tokens_wasted_, 1);
	link_int(0, "compliant", &compliant_, 1);
	link_int(0, "non-compliant", &non_compliant_,1);
}

void NewRenoPoliceTcpAgent::reset()
{
	TcpAgent::reset();

        last_arriv_ = 0;
        tokens_avail_ = 0;
	tokens_wasted_ = 0;
	compliant_ = 0;
	non_compliant_ = 0;
	lbdepth_bytes_ = (lbrate_/8) * lbdepth_;
}

void NewRenoPoliceTcpAgent::update_bucket()
{
	double now = Scheduler::instance().clock();

	// Recalculation is to allow you to change the lbdepth on the fly
	// Yes, I know this is inefficient.

	lbdepth_bytes_ = (lbrate_/8) * lbdepth_;

	// update tokens
        tokens_avail_ += double((now-last_arriv_)*lbrate_)/8.0;
        last_arriv_ = now;
        if (tokens_avail_ > lbdepth_bytes_ ) {
                tokens_wasted_ += (tokens_avail_ - lbdepth_bytes_ );
                tokens_avail_ = lbdepth_bytes_ ;
		if (trace_) {
			sprintf(trace_->buffer(),"loss %g %5.0f",now,tokens_wasted_ );
			trace_->dump();
		}
        }

}

/*
 * Called from DataSource
 */
void NewRenoPoliceTcpAgent::output(int seqno, int reason)
{
	Packet* p = allocpkt(seqno);
	double now = Scheduler::instance().clock();

	update_bucket();

        // check to see if packet is compliant
        // mark packet if compliant
        if (p->size_ <= tokens_avail_) {
		p->flags_ |= PF_PRI;
		tokens_avail_ = tokens_avail_ - p->size_;
		compliant_++;	
        } else {
		non_compliant_++;
	}
        
	p->bd_.tcp_.ts_ = now;
	p->flags_ |= reason;
	node_->transmit(p);
	if (seqno > maxseq_) {
		maxseq_ = seqno;
		if (!rtt_active_) {
			rtt_active_ = 1;
			if (seqno > rtt_seq_)
				rtt_seq_ = seqno;
		}
	}
	if (!pending_[TCP_TIMER_RTX])
		/* No timer pending.  Schedule one. */
		set_rtx_timer();
}

void NewRenoPoliceTcpAgent::send(int force, int reason, int maxburst)
{
        int win = window();
        int npackets = 0;

	update_bucket();
        while (seqno_ <= highest_ack_ + win && seqno_ < curseq_) {
                if (overhead_ == 0 || force || ( tokens_avail_ > size_ ) || pending_[TCP_TIMER_DELSND]) {
                        output(seqno_++, reason);
                        npackets++;
                } else if (!pending_[TCP_TIMER_DELSND]) {
                        /*
                         * Set a delayed send timeout.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
                }
                if (maxburst && npackets == maxburst)
                        break;
        }
}

void NewRenoPoliceTcpAgent::plot() 
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "%g %f %d %f %d", t, cwnd_, last_ack_, cwnd_ - seqno_ + last_ack_ , seqno_ );
	trace_->dump();
} 


class NewRenoCloadTcpAgent : virtual public NewRenoPoliceTcpAgent {
public:
	NewRenoCloadTcpAgent();
	virtual void closecwnd(int how);
	virtual void opencwnd();
protected:
};

static class NewRenoCloadTcpMatcher : public Matcher {
public:
	NewRenoCloadTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno-cload") == 0)
			return (new NewRenoCloadTcpAgent());
		return (0);
	}
} matcher_renocload;

NewRenoCloadTcpAgent::NewRenoCloadTcpAgent()
{
}

/*
 * open up the congestion window
 */
void NewRenoCloadTcpAgent::opencwnd()
{
	double rbcwnd, prorate, above_rbcwnd;

	switch (wnd_option_) {
	case 0:
		rbcwnd = double( lbrate_ * t_brtt_ * tcp_tick_ / (8*size_) );
		break;
	case 1:
	case 2:
		rbcwnd = double( lbrate_ * t_rtt_ * tcp_tick_ / (8*size_) );
		break;
	default:
		abort();
	}

	if (cwnd_ < (rbcwnd + 1)) {
		cwnd_ = rbcwnd + 1;
		above_rbcwnd = 1;
		return;
	}
	else {
		above_rbcwnd = cwnd_ - rbcwnd;
		prorate = double( above_rbcwnd / cwnd_ );
	}

	if (cwnd_ < ssthresh_) {
		switch (wnd_option_) {
		case 0:
		case 1:
			cwnd_ += prorate;
			break;
		case 2:
			/* slow-start (exponential) */
			cwnd_ += 1;
			break;
		}
	} else {
		/* linear */
		switch (wnd_option_) {
		case 0:
		case 1:
		case 2:
			/* This is the standard algorithm. */
			cwnd_ += 1 / cwnd_;
			break;

		default:
#ifdef notdef
			/*XXX*/
			error("illegal window option %d", wnd_option_);
#endif
			abort();
		}
	}
	// if maxcwnd_ is set (nonzero), make it the cwnd limit
	if (maxcwnd_ && (int(cwnd_) > maxcwnd_))
		cwnd_ = maxcwnd_;

	return;
}

/*
 * close down the congestion window
 */
void NewRenoCloadTcpAgent::closecwnd(int how)
{
        double rbcwnd, above_rbcwnd;  /* rate based cwnd */

	switch (wnd_option_) {
	case 0:
		rbcwnd = double( lbrate_ * t_brtt_ * tcp_tick_ / (8*size_) );
		break;
	case 1:
	case 2:
		rbcwnd = double( lbrate_ * t_rtt_ * tcp_tick_ / (8*size_) );
		break;
	default:
		abort();
	}

	if ( cwnd_ < rbcwnd ) {
		cwnd_ = rbcwnd + 1;
		return;
	}
	else {
		above_rbcwnd = cwnd_ - rbcwnd ;
		switch (how) {
		case 0:
			ssthresh_ = int(rbcwnd + above_rbcwnd/2);
			cwnd_ = rbcwnd + 1;
			break;
			/* timeout: set cwnd to rate based window.
			*          set ssthresh to rbwnd+(cwnd-rbwnd)/2
			*/
		case 1:
			cwnd_ = rbcwnd + above_rbcwnd/2;
			ssthresh_ = int(cwnd_);
			break;
		case 2:
			/* Tahoe dup acks: set cwnd to rate based window.
			* No congestion avoidance (ssthresh) */
			cwnd_ = rbcwnd + 1;
			break;
		default:
			abort();
		}
	}
	fcnt_ = 0.;
	count_ = 0;
}

class NewRenoTimedTcpAgent : virtual public NewRenoPoliceTcpAgent {
public:
	NewRenoTimedTcpAgent();
	virtual void timeout(int tno);
	virtual void reset();
	void send(int force, int reason, int maxburst = 0);
protected:
	void set_rbsnd_timer();
};

static class NewRenoTimedTcpMatcher : public Matcher {
public:
	NewRenoTimedTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno-timed") == 0)
			return (new NewRenoTimedTcpAgent());
		return (0);
	}
} matcher_renotimed;

NewRenoTimedTcpAgent::NewRenoTimedTcpAgent()
{
}

void NewRenoTimedTcpAgent::reset()
{
	NewRenoPoliceTcpAgent::reset();
	set_rbsnd_timer();
}

void NewRenoTimedTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
	}

	/* retransmit timer */
	if (tno == TCP_TIMER_RTX) {
		recover_ = maxseq_;
		recover_cause_ = 2;
		closecwnd(0);
		reset_rtx_timer(0);
		send(0, PF_TIMEOUT);
	} else if (tno == TCP_TIMER_DELSND) {
		/*
		 * delayed-send timer, with random overhead
		 * to avoid phase effects
		 */
		send(1, PF_TIMEOUT);
	} else if (tno == TCP_TIMER_RBSND) {
		update_bucket();
		if (trace_) {
			sprintf(trace_->buffer(), "t %g tok=%g",
				Scheduler::instance().clock(), tokens_avail_);
			trace_->dump();
		}
		send(1, PF_RBSND);
		if (pending_[TCP_TIMER_RBSND]) {
		  fprintf(stderr,"rbsnd timer should have been cancelled\n");
		  abort();
		}
		set_rbsnd_timer();
	}

}

/*
 * Try to send as much data as the window will allow.  The link layer will
 * do the buffering; we ask the application layer for the size of the packets.
 */
void NewRenoTimedTcpAgent::send(int force, int reason, int maxburst)
{
        int win = window();
        int npackets = 0;

	update_bucket();

	if (reason == PF_RBSND) {
        	while ( (tokens_avail_ > size_) && 
			(seqno_ <= last_ack_ + wnd_) && (seqno_ < curseq_) ) {
			output(seqno_++, reason);
			npackets++;
			if (maxburst && npackets == maxburst)
				break;
		}
	} else {
        	while ( (seqno_ <= highest_ack_ + win) && seqno_ < curseq_) {
                	if (overhead_ == 0 || force || (tokens_avail_ > size_) || pending_[TCP_TIMER_DELSND]) {
                        	output(seqno_++, reason);
                        	npackets++;
                	} else if (!pending_[TCP_TIMER_DELSND]) {
                        	/*
                         	* Set a delayed send timeout.
                         	*/
                        	sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        	return;
                	}
                if (maxburst && npackets == maxburst)
                        break;
		}
        }
}

void NewRenoTimedTcpAgent::set_rbsnd_timer()
{
	if (pending_[TCP_TIMER_RBSND]) {
		cancel(TCP_TIMER_RBSND);
	}
	sched(timer_interval_,TCP_TIMER_RBSND);
}

class NewRenoTimedCloadTcpAgent : public NewRenoTimedTcpAgent, public NewRenoCloadTcpAgent {
public:
	NewRenoTimedCloadTcpAgent();
protected:
};

static class NewRenoTimedCloadTcpMatcher : public Matcher {
public:
	NewRenoTimedCloadTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno-timed-cload") == 0)
			return (new NewRenoTimedCloadTcpAgent());
		return (0);
	}
} matcher_renotimedcload;

NewRenoTimedCloadTcpAgent::NewRenoTimedCloadTcpAgent()
{
}
