#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"
#include "trace.h"
#include "random.h"
#include "node.h"
#include "scoreboard.h"

#define MAXBINS 1024
#define TCP_INDEPENDENT 0
#define TCP_AWARE 1
#define TCP_MODIFIED 2
#define TCP_MODADAPT 3

class NewRenoPriorityTcpAgent : public TcpAgent {
public:
	NewRenoPriorityTcpAgent();
	virtual int window();
	virtual void output(int seqno, int reason);
	virtual void reset();
	void send(int force, int reason, int maxburst = 0);
	virtual void closecwnd(int how);
	virtual void opencwnd();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
	virtual void plot();
protected:
	u_int dupwnd_;
	int changes_;		/* 0 for fixing unnecessary fast retransmits */
				/* 1 for additional code from Allman, */
				/*  to implement other algorithms from */
				/*  Hoe's paper */
	void partialnewack(Packet *pkt);
	void set_bwcheck_timer();
	void set_probe_timer();
	int marked_;
	int unmarked_;
	int algorithm_;
	int history_;
        int progress_cnt_;
        int seqno_progress_[MAXBINS];
        int pssthresh_;
        int bssthresh_;
        int bins_;
	int do_mark_;
	int m_pkts_;
	int m_interdrop_;
	int u_pkts_;
	int u_interdrop_;
	int probe_cnt_;
	int probe_ival_;
	int probe_max_;
	double wbwa_;
	double sbwa_;
	double bwa_;
	double weight_;
	double mark_cnt_;
	double target_bw_;
	double bwcheck_interval_;
	double probe_interval_;
	double probability_;
	double increment_;
	PriorityScoreBoard pscb_;
private:
	int markpacket();
	double status();
	void update_status();
};

static class NewRenoPriorityTcpMatcher : public Matcher {
public:
	NewRenoPriorityTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno-priority") == 0)
			return (new NewRenoPriorityTcpAgent());
		return (0);
	}
} matcher_renopriority;

NewRenoPriorityTcpAgent::NewRenoPriorityTcpAgent() : dupwnd_(0), changes_(0)
{
        link_bw("ns_tcp_newreno_priority", "target-bw", &target_bw_, 0);
        link_int("ns_tcp_newreno_priority", "algorithm", &algorithm_, 0);
        link_int("ns_tcp_newreno_priority", "history", &history_, 0);
        link_real("ns_tcp_newreno_priority", "weight", &weight_, 0);
	link_real("ns_tcp_newreno_priority", "increment", &increment_, 0);
        link_time("ns_tcp_newreno_priority", "bwcheck-interval", &bwcheck_interval_, 0);
        link_time("ns_tcp_newreno_priority", "probe-interval", &probe_interval_, 0);
        link_int("ns_tcp_newreno_priority", "probe-max", &probe_max_, 0);
	link_int("ns_tcp_newreno_priority", "changes", &changes_, 0);
	link_int("ns_tcp_newreno_priority", "bins", &bins_,0);

        link_int(0, "do-mark", &do_mark_, 1);
        link_int(0, "marked", &marked_, 1);
        link_int(0, "unmarked", &unmarked_,1);
        link_int(0, "bssthresh", &bssthresh_, 1);
        link_int(0, "pssthresh", &pssthresh_, 1);
        link_real(0, "wbwa", &wbwa_, 1);
        link_real(0, "sbwa", &sbwa_, 1);
        link_real(0, "bwa", &bwa_, 1);
        link_real(0, "probability", &probability_, 1);
}

void NewRenoPriorityTcpAgent::reset()
{
	TcpAgent::reset();

	m_pkts_ = 0;
	m_interdrop_ = 0;
	u_pkts_ = 0;
	u_interdrop_ = 0;
	do_mark_ = 1;
	probe_cnt_ = 0;
	probe_ival_ = 1;
	wbwa_ = 0.99*target_bw_;
	sbwa_ = 0.99*target_bw_;
	bwa_ = 0.99*target_bw_;
	marked_ = 0;
	unmarked_ = 0;
	mark_cnt_ = 1;
	// probability_ = 1;
	progress_cnt_ = 0;

	for (int i=0; i < MAXBINS; i++) {
	  seqno_progress_[i] = 0;
	}
	
	set_bwcheck_timer();
	set_probe_timer();
}

int NewRenoPriorityTcpAgent::window()
{
        //
        // reno: inflate the window by dupwnd_
        //      dupwnd_ will be non-zero during fast recovery,
        //      at which time it contains the number of dup acks
        //
	int win = int(cwnd_) + dupwnd_;
        if (win > int(wnd_))
                win = int(wnd_);
        return (win);
}

/* 
 * Process a packet that acks previously unacknowleges data, but
 * does not take us out of Fast Retransmit.
 */
void NewRenoPriorityTcpAgent::partialnewack(Packet* pkt)
{
        newtimer(pkt);
#ifdef notyet
        if (pkt->seqno_ == stp->maxpkts && stp->maxpkts > 0)
                stp->endtime = (float) realtime();
#endif
        last_ack_ = pkt->seqno_;
        highest_ack_ = last_ack_;
        if (seqno_ < last_ack_ + 1)
                seqno_ = last_ack_ + 1;
        if (rtt_active_ && pkt->seqno_ >= rtt_seq_) {
                rtt_active_ = 0;
                t_backoff_ = 1;
        }
}

void NewRenoPriorityTcpAgent::recv(Packet *pkt)
{
	static int acked = 0;
	int new_ssthresh;
	static double ack2, ack3, basertt;

        /* Use first packet to calculate the RTT  --contributed by Allman */

        if (++acked == 1)
                basertt = Scheduler::instance().clock() - firstsent_;

        /* Estimate ssthresh based on the calculated RTT and the estimated
           bandwidth (using ACKs 2 and 3).  */

        else if (acked == 2)
                ack2 = Scheduler::instance().clock();
        else if (acked == 3)
        {       
                ack3 = Scheduler::instance().clock();  
                new_ssthresh = int((basertt * (size_ / (ack3 - ack2)))
                                    / size_);
                if (changes_ > 0 && new_ssthresh < ssthresh_)
                    ssthresh_ = new_ssthresh;
        }

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
	if (pkt->seqno_ > last_ack_) {
	    if (pkt->seqno_ >= recover_) {
		dupwnd_ = 0;
		newack(pkt);
                opencwnd();
	    }
	    else {
		/* received new ack for a packet sent during Fast
		 *  Recovery, but sender stays in Fast Recovery */
		dupwnd_ = 0;
		partialnewack(pkt);
		output(last_ack_ + 1, 0);
	    }
   	} else if (pkt->seqno_ == last_ack_)  {
		if (++dupacks_ == NUMDUPACKS) {
			/*
			 * Assume we dropped just one packet.
			 * Retransmit last ack + 1
			 * and try to resume the sequence.
			 */
                       /* The line below, for "bug_fix_" true, avoids
                        * problems with multiple fast retransmits after
			* a retransmit timeout.
                        */
			if ( (highest_ack_ > recover_) ||
			    ( recover_cause_ != 2)) {
				recover_cause_ = 1;
				recover_ = maxseq_;
				switch (algorithm_) {
				case TCP_MODIFIED:
				case TCP_MODADAPT:
				  if (pscb_.StatPacket(last_ack_ + 1)) {
				    marked_--;
				    m_interdrop_ = m_pkts_;
				    m_pkts_ = 0;
				  }
				  else {
				    unmarked_--;
				    u_interdrop_ = u_pkts_;
				    u_pkts_ = 0;
				  }
				  if (pscb_.StatPacket(last_ack_ + 1))
				    closecwnd(1);
				  else
				    closecwnd(3);
				  break;
				case TCP_INDEPENDENT:
				case TCP_AWARE:
				default:
				  closecwnd(1);
				  break;
				}
				reset_rtx_timer(1);
				output(last_ack_ + 1, PF_DUPACK);
                        }
			dupwnd_ = NUMDUPACKS;
		} else if (dupacks_ > NUMDUPACKS) 
			++dupwnd_;
                        /* For every two duplicate ACKs we receive (in the
                         * "fast retransmit phase"), send one entirely new
                         * data packet "to keep the flywheel going".  --Allman 
			 */
			 if (changes_ > 0 && (dupacks_ % 2) == 1)
		   		output (seqno_++,0);

	}
	Packet::free(pkt);
	// if (trace_)
		// plot();

	/*
	 * Try to send more data
	 */

	if (dupacks_ == 0) 
                /*
                 * Maxburst is really only needed for the first
                 *   window of data on exiting Fast Recovery.
                 */
		send(0, 0, maxburst_);
	else if (dupacks_ > NUMDUPACKS - 1 && changes_==0)
		send(0, 0, 2);
}

void NewRenoPriorityTcpAgent::update_status()
{
	double bw, bw_tmp = 0;
	int packets,i,seqno_var;
	double now = Scheduler::instance().clock();

	if ((seqno_ >= curseq_) && (last_ack_ + 1 == seqno_)) {
	  // Don't update status if application is idle
	  return;
	}
	switch (history_) {
	case 0:
		// no history
	        break;
	case 1:		
	        // sliding window
	        if (bins_ > MAXBINS) bins_=MAXBINS;
	        seqno_progress_[progress_cnt_ % bins_] = last_ack_;
		progress_cnt_++;

		// This is a cirular array so progress_cnt_++
		// entry contains the seqno at the beginning of epoch
		seqno_var = seqno_progress_[progress_cnt_ % bins_];
		sbwa_ = (last_ack_ - seqno_var)*size_*8/(bins_*bwcheck_interval_);
		break;
	case 2:
		// weighted average
	        if (bins_ > MAXBINS) bins_=MAXBINS;
	        seqno_progress_[progress_cnt_ % bins_] = last_ack_;
		seqno_var = seqno_progress_[(progress_cnt_ -1) % bins_];
		progress_cnt_++;
		
		packets = last_ack_ - seqno_var;
		bw = (double) packets*size_*8/bwcheck_interval_;
		wbwa_ = weight_ * bw + (1-weight_)*wbwa_;
		break;
		  
	case 3: 
		// entire history
		// NOT GOOD WHEN APP IDLE
		bwa_ = (last_ack_*size_*8)/now;
		break;
	}
}

double NewRenoPriorityTcpAgent::status()
{
	double tcwnd,metric;
	switch (history_) {
		case 0:
		  // no history
		  tcwnd = target_bw_*t_rtt_*tcp_tick_;
		  metric = (cwnd_ - tcwnd)/tcwnd;
		  break;
		case 1:
		  // sliding window
		  metric = (sbwa_ - target_bw_)/target_bw_;
		  break;
		case 2:
		  // weighted average
		  metric = (wbwa_ - target_bw_)/target_bw_;
		  break;
		case 3:
		  // entire history
		  metric = (bwa_ - target_bw_)/target_bw_;
		  break;
	}
	return(metric);
}

void NewRenoPriorityTcpAgent::timeout(int tno)
{
	double pwnd;
	double now = Scheduler::instance().clock();
	double stat;

	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
	} 
	if (tno == TCP_TIMER_BWCHECK) {
		update_status();
		switch (algorithm_) {
		case TCP_INDEPENDENT:
		  // NO MODS TO TCP
		  if (do_mark_) {
		    stat = status();
		    if (stat < 0) {
		      probability_ += -stat*increment_;
		      if (probability_ > 1) probability_ = 1;
		    }
		    else {
		      probability_ -= stat*increment_;
		      if (probability_ < 0) probability_ = 0;
		    }
		  }
		  break;
		case TCP_AWARE:
		  // NO MODS TO TCP, BUT HAS INFO ON TCP CONNECTION
		  if (do_mark_) {
		    stat = status();
		    if (stat < 0) {
		      pwnd = probability_ * cwnd_;
		      pwnd += bwcheck_interval_/(t_rtt_*tcp_tick_);
		      probability_ = pwnd / cwnd_;
		      if (probability_ > 1) probability_ = 1;
		    }
		    else {
		      pwnd = probability_ * cwnd_;
		      pwnd -= bwcheck_interval_/(t_rtt_*tcp_tick_);
		      probability_ = pwnd / cwnd_;
		      if (probability_ < 0) probability_ = 0;
		    }
		  }
		  break;
		case TCP_MODIFIED:
		case TCP_MODADAPT:
		  // MODS TO TCP WINDOWING
		  // Don't need to do anything except update status
		  break;
		default:
		  break;
		}
		set_bwcheck_timer();
	}
	if (tno == TCP_TIMER_PROBE) {
	  if (!do_mark_) {
	    probe_cnt_--;
	    if (probe_cnt_ == 0) {
	      do_mark_=1;
	    }
	  }
	  else
	    if ((m_interdrop_ > 0) && (u_interdrop_ > 0)) {
	      if (m_interdrop_ < 10*u_interdrop_) {
		m_interdrop_ = 0;
		m_pkts_ = 0;
		u_interdrop_ = 0;
		u_pkts_ = 0;
		probability_ = 0;
		do_mark_ = 0;
		probe_cnt_ = probe_ival_;
		probe_ival_ = 2*probe_ival_;
		if (probe_ival_ > probe_max_) probe_ival_ = probe_max_;
	      }
	      else {
		probe_ival_ = 1;
	      }
	    }
	  set_probe_timer();
	}
	TcpAgent::timeout(tno);
}

int NewRenoPriorityTcpAgent::markpacket()
{
	mark_cnt_ += probability_;

	if (!do_mark_) return(0);
	if (mark_cnt_ > 1) {
		mark_cnt_--;
		return(1);
	}
	else {
		return(0);
	}
}

/*
 * Called from DataSource
 */
void NewRenoPriorityTcpAgent::output(int seqno, int reason)
{
	Packet* p = allocpkt(seqno);
	double now = Scheduler::instance().clock();

	if (trace_)
		plot();
        if (markpacket()) {
		p->flags_ |= PF_PRI;
		marked_++;
		m_pkts_++;
		pscb_.LogPacket(seqno, 1); // UPDATE SCOREBOARD
        } else {
		unmarked_++;
		u_pkts_++;
		pscb_.LogPacket(seqno, 0); // UPDATE SCOREBOARD
	}
        
	p->bd_.tcp_.ts_ = now;
	p->flags_ |= reason;
	node_->transmit(p);
	if (seqno > maxseq_) {
		maxseq_ = seqno;
		if (!rtt_active_) {
			rtt_active_ = 1;
			if (seqno > rtt_seq_)
				rtt_seq_ = seqno;
		}
	}
	if (!pending_[TCP_TIMER_RTX])
		/* No timer pending.  Schedule one. */
		set_rtx_timer();
}

void NewRenoPriorityTcpAgent::send(int force, int reason, int maxburst)
{
        int win = window();
        int npackets = 0;

        while (seqno_ <= highest_ack_ + win && seqno_ < curseq_) {
                if (overhead_ == 0 || force || pending_[TCP_TIMER_DELSND]) {
                        output(seqno_++, reason);
                        npackets++;
                } else if (!pending_[TCP_TIMER_DELSND]) {
                        /*
                         * Set a delayed send timeout.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
                }
                if (maxburst && npackets == maxburst)
                        break;
        }
}

/*
 * open up the congestion window
 */
void NewRenoPriorityTcpAgent::opencwnd()
{
  double pwnd,bwnd,stat;
  switch (algorithm_) {
  case TCP_INDEPENDENT:
  case TCP_AWARE:
    TcpAgent::opencwnd();
    break;
  case TCP_MODADAPT:
  case TCP_MODIFIED:
    if (do_mark_) {
	pwnd = probability_ * cwnd_;
	bwnd = cwnd_ - pwnd;
	stat = status();
	if (stat > 0) {
		// I HAVE ENOUGH BANDWIDTH
		if (pwnd > 0) {
		        // Decrease pwnd and keep cwnd constant.  
		        // This grows bwnd so only decrease pwnd by amount 
		        // allowed for bwnd
			if (bwnd < bssthresh_)
				pwnd -= bwnd/cwnd_;
			else
				pwnd -= 1/cwnd_;
		}
		else {
		        // Increase bwnd (= cwnd) by amount allowed for bwnd
		        if (bwnd < bssthresh_)
			        cwnd_ += 1;
			else
			        cwnd_ += 1/cwnd_;
		}
		if (pwnd < 0)
			pwnd = 0;
		probability_ = pwnd / cwnd_;
	}
	else {
		// I NEED MORE BANDWIDTH!
		if ((pwnd < pssthresh_) && (bwnd < bssthresh_)) {
			// Both pwnd and bwnd are slow-startable
			// cwnd_ grows by cwnd_ every round trip time
			// RARELY INVOKED
			pwnd += pwnd/cwnd_;
			bwnd += bwnd/cwnd_;
		}
		else if (pwnd < pssthresh_) {
			// Only pwnd is slow startable
			// cwnd_ grows by (pwnd + 1) every round trip time
			// SOMETIMES INVOKED
			pwnd += pwnd/cwnd_;
			bwnd += 1/cwnd_;
		}
		else if (bwnd < bssthresh_) {
			// Only bwnd is slow startable
			// cwnd_ grows by (bwnd + 1) every round trip time
			// RARELY INVOKED
			pwnd += 1/cwnd_;
			bwnd += bwnd/cwnd_;
		} else {
			// Neither are slow startable
			// cwnd_ grows by 2 every round trip time
			// INVOKED OFTEN
			pwnd += 1/cwnd_;
			bwnd += 1/cwnd_;
		}
		cwnd_ = pwnd + bwnd;
		probability_ = pwnd / cwnd_;
	}
    }
    else 
      TcpAgent::opencwnd();
    break;
  default:
    TcpAgent::opencwnd();
  }
}

/*
 * close down the congestion window
 */
void NewRenoPriorityTcpAgent::closecwnd(int how)
{
  double pwnd, bwnd;
  switch (algorithm_) {
  case TCP_INDEPENDENT:
  case TCP_AWARE:
    TcpAgent::closecwnd(how);
    break;
  case TCP_MODIFIED:
  case TCP_MODADAPT:
      pwnd = probability_ * cwnd_;
      bwnd = cwnd_ - pwnd;
      switch (how) {
      case 0:
	ssthresh_ = int(window()/2);
	cwnd_ = int(wnd_init_);
	probability_ = 1;
	break;
      case 1:
	cwnd_ = window()/2;
	ssthresh_ = int(cwnd_);
	break;
      case 2:
	cwnd_ = wnd_init_;
	break;
      case 3:
	if (do_mark_) {
	  bwnd = bwnd/2;
	  cwnd_ = pwnd + bwnd;
	  ssthresh_ = int(cwnd_);
	  probability_ = pwnd / cwnd_;
	} else {
	  cwnd_ = window()/2;
	  ssthresh_ = int(cwnd_);
	}
	break;
      default:
	abort();
      }
      pssthresh_ = int(probability_*ssthresh_);
      bssthresh_ = int((1-probability_)*ssthresh_);
      count_ = 0;
      break;
  default:
    TcpAgent::closecwnd(how);
  }
}

void NewRenoPriorityTcpAgent::set_bwcheck_timer()
{
	if (pending_[TCP_TIMER_BWCHECK]) {
		cancel(TCP_TIMER_BWCHECK);
	}
	sched(bwcheck_interval_,TCP_TIMER_BWCHECK);
}

void NewRenoPriorityTcpAgent::set_probe_timer()
{
	if (pending_[TCP_TIMER_PROBE]) {
		cancel(TCP_TIMER_PROBE);
	}
	sched(probe_interval_,TCP_TIMER_PROBE);
}

void NewRenoPriorityTcpAgent::plot()
{
	double now = Scheduler::instance().clock();
	if (trace_) {
		sprintf(trace_->buffer(),"%g %d %d %5.2f %5.2f %5.2f",now,marked_,unmarked_, probability_*cwnd_, (1-probability_)*cwnd_, cwnd_);
		trace_->dump();
	}
}
