#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"
#include "node.h"
#include "scoreboard.h"
#include "random.h"
#include "trace.h"

#define TRUE    1
#define FALSE   0
#define RECOVER_DUPACK  1
#define RECOVER_TIMEOUT 2
#define RECOVER_QUENCH  3

class Sack1PriorityTcpAgent : public TcpAgent {
 public:
	Sack1PriorityTcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
	virtual void output(int seqno, int reason);
	virtual void reset();
	virtual void closecwnd(int how);
	virtual void opencwnd();
	void plot();
	void send(int force, int reason, int maxburst = 0);
	int markpacket();
 protected:
	int marked_;
	int unmarked_;
	int greedy_;
	int algorithm_;
	int prev_last_ack_;
	double wbwa_;
	double weight_;
	double prev_time_;
	double pwnd_;
	double bwnd_;
	double mark_cnt_;
	double target_bw_;
	double bwcheck_interval_;
	double probability_;
	PriorityScoreBoard pscb_;
	void set_bwcheck_timer();
	u_char timeout_;	/* boolean: sent pkt from timeout? */
	u_char fastrecov_;      /* boolean: doing fast recovery? */
	int pipe_;	      /* estimate of pipe size (fast recovery) */ 
	ScoreBoard scb_;
};

static class Sack1PriorityTcpMatcher : public Matcher {
public:
	Sack1PriorityTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack1-priority") == 0)
			return (new Sack1PriorityTcpAgent());
		return (0);
	}
} matcher_sackpriority;

int Sack1PriorityTcpAgent::window()
{
	return(int(cwnd_ < wnd_ ? cwnd_ : wnd_));
}

Sack1PriorityTcpAgent::Sack1PriorityTcpAgent() : pipe_(-1), fastrecov_(FALSE)
{
	link_bw("ns_tcp_sack1_priority", "target-bw", &target_bw_, 0);
	link_int("ns_tcp_sack1_priority", "greedy", &greedy_, 0);
	link_int("ns_tcp_sack1_priority", "algorithm", &algorithm_, 0);
	link_real("ns_tcp_sack1_priority", "weight", &weight_, 0);
	link_time("ns_tcp_sack1_priority", "bwcheck-interval", &bwcheck_interval_, 0);

	link_int(0, "marked", &marked_, 1);
	link_int(0, "unmarked", &unmarked_,1);
	link_int(0, "prev-last-ack", &prev_last_ack_,1);
	link_real(0, "pwnd", &pwnd_, 1);
	link_real(0, "bwnd", &bwnd_,1);
	link_real(0, "wbwa", &wbwa_, 1);
	link_real(0, "prev-time", &prev_time_, 1);
	link_real(0, "probability", &probability_, 1);
}

void Sack1PriorityTcpAgent::recv(Packet *pkt)
{
	int xmit_seqno;
	int dontSend = 0;

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
	if (!fastrecov_) {
		/* normal... not fast recovery */
		if ((int)pkt->seqno_ > last_ack_) {
			/*
			 * regular ACK not in fast recovery... normal
			 */
			newack(pkt);
			opencwnd();
			timeout_ = FALSE;
			scb_.ClearScoreBoard();
		} else if (timeout_ == FALSE)  {
			if (pkt->seqno_ != last_ack_) {
			        fprintf(stderr, "pkt seq %d should be %d\n" ,
			                pkt->seqno_, last_ack_);
			        abort();
			}
			scb_.UpdateScoreBoard (last_ack_, pkt);
			/*
			 * a duplicate ACK
			 */
			if (++dupacks_ == NUMDUPACKS) {
			        /*
			         * Assume we dropped just one packet.
			         * Retransmit last ack + 1
			         * and try to resume the sequence.
			         */
			        if ((highest_ack_ > recover_) ||
			            (recover_cause_ != RECOVER_TIMEOUT)) {
			                recover_cause_ = RECOVER_DUPACK;
			                recover_ = maxseq_;
			                pipe_ = int(cwnd_) - NUMDUPACKS;
			                closecwnd(1);
			                reset_rtx_timer(1);
					fastrecov_ = TRUE;
					scb_.MarkRetran (last_ack_+1);
			                output(last_ack_ + 1, PF_DUPACK);
			        }
			}
		}
		if (dupacks_ == 0)
			send(FALSE, 0, maxburst_);
	} else {
		/* we are in fast recovery */
		--pipe_;
		if ((int)pkt->seqno_ >= recover_) {
			/* ACK indicates fast recovery is over */
			recover_ = 0;
			fastrecov_ = FALSE;
			newack(pkt);
			timeout_ = FALSE;
			scb_.ClearScoreBoard();

			/* New window: W/2 - K or W/2? */
		} else if ((int)pkt->seqno_ > highest_ack_) {
			/* Not out of fast recovery yet.
			 * Update highest_ack_, but not last_ack_. */
			--pipe_;
			highest_ack_ = (int)pkt->seqno_;
		 	scb_.UpdateScoreBoard (highest_ack_, pkt);
			t_backoff_ = 1;
			newtimer(pkt);
		} else if (timeout_ == FALSE) {
			/* got another dup ack */
			scb_.UpdateScoreBoard (last_ack_, pkt);
			if (dupacks_ > 0)
			        dupacks_++;
		}
		send(FALSE, 0, maxburst_);
	}

	Packet::free(pkt);
	if (trace_)
		plot();
}

void Sack1PriorityTcpAgent::send(int force, int reason, int maxburst)
{
	register int pktno, found, nextpktno, npacket = 0;
	int win = window();
	int xmit_seqno;

	found = 1;
	/*
	 * as long as the pipe is open and there is app data to send...
	 */
	while (((!fastrecov_  && (seqno_ <= last_ack_ + win)) ||
	    (fastrecov_ && (pipe_ < int(cwnd_)))) 
		&& seqno_ < curseq_ && found) {

		if (overhead_ == 0 || force) {
			found = 0;

			xmit_seqno = scb_.GetNextRetran ();
#ifdef DEBUGSACK1A
			printf("highest_ack: %d xmit_seqno: %d\n", 
			highest_ack_, xmit_seqno);
#endif
			if (xmit_seqno == -1) { 
			    if ((!fastrecov_ && seqno_<=highest_ack_+win)||
			     (fastrecov_ && seqno_<=last_ack_+int(2*wnd_))){ 
				found = 1;
				xmit_seqno = seqno_++;
#ifdef DEBUGSACK1A
				printf("sending %d fastrecovery: %d win %d\n",
				  xmit_seqno, fastrecov_, win);
#endif
			    }
			}
			else if (recover_>0 &&
				xmit_seqno<=last_ack_+int(2*wnd_)) {
				found = 1;
				scb_.MarkRetran (xmit_seqno);
				win = window();
			}
			if (found) {
				output(xmit_seqno, reason);
			        if (seqno_ <= xmit_seqno)
			                seqno_ = xmit_seqno + 1;
				npacket++;
				pipe_++;
			}
		} else if (!pending_[TCP_TIMER_DELSND]) {
			/*
			 * Set a delayed send timeout.
			 * This is only for the simulator,to add some
			 *   randomization if speficied.
			 */
			sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
			return;
		}
		if (maxburst && npacket == maxburst)
			break;
	} /* while */
}

void Sack1PriorityTcpAgent::plot()
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "%g %f %d %f %d", t, cwnd_, last_ack_, cwnd_ -seqno_ + last_ack_ , seqno_ );
	trace_->dump();
}

int Sack1PriorityTcpAgent::markpacket()
{
	mark_cnt_ += pwnd_/cwnd_;
	if (mark_cnt_ > 1) {
		mark_cnt_ -= floor(mark_cnt_);
		return(1);
	}
	else {
		return(0);
	}
}

/*
 * Called from DataSource
 */
void Sack1PriorityTcpAgent::output(int seqno, int reason)
{
	Packet* p = allocpkt(seqno);
	double now = Scheduler::instance().clock();

	if (markpacket()) {
		p->flags_ |= PF_PRI;
		marked_++;
		pscb_.LogPacket(seqno, 1); // UPDATE SCOREBOARD
	} else {
		unmarked_++;
		pscb_.LogPacket(seqno, 0); // UPDATE SCOREBOARD
	}
	
	p->bd_.tcp_.ts_ = now;
	p->flags_ |= reason;
	node_->transmit(p);
	if (seqno > maxseq_) {
		maxseq_ = seqno;
		if (!rtt_active_) {
			rtt_active_ = 1;
			if (seqno > rtt_seq_)
				rtt_seq_ = seqno;
		}
	}
	if (!pending_[TCP_TIMER_RTX])
		/* No timer pending.  Schedule one. */
		set_rtx_timer();
}

/*
 * open up the congestion window
 */
void Sack1PriorityTcpAgent::opencwnd()
{
	TcpAgent::opencwnd();
}

/*
 * close down the congestion window
 */
void Sack1PriorityTcpAgent::closecwnd(int how)
{
	TcpAgent::closecwnd(how);
}

void Sack1PriorityTcpAgent::reset()
{
	TcpAgent::reset();

	prev_time_ = 0;
	prev_last_ack_ = 0;
	wbwa_ = 0;
	marked_ = 0;
	unmarked_ = 0;
	pwnd_ = cwnd_;
	bwnd_ = 0;
	mark_cnt_ = 1;
	probability_ = 1;
	
	set_bwcheck_timer();
}

void Sack1PriorityTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupacks_ = 0;
		fastrecov_ = FALSE;
		timeout_ = TRUE;
		if (highest_ack_ > last_ack_)
			last_ack_ = highest_ack_;
#ifdef DEBUGSACK1A
		printf ("timeout. highest_ack: %d seqno: %d\n", 
		  highest_ack_, seqno_);
#endif
		recover_ = maxseq_;
		scb_.ClearScoreBoard();
	}
		/* retransmit timer */
	if (tno == TCP_TIMER_RTX) {
		recover_ = maxseq_;
		recover_cause_ = 2;
		closecwnd(0);
		reset_rtx_timer(0);
		send(0, PF_TIMEOUT);
	} else if (tno == TCP_TIMER_DELSND) {
		/*
		 * delayed-send timer, with random overhead
		 * to avoid phase effects
		 */
		send(1, PF_TIMEOUT);
	} else if (tno == TCP_TIMER_BWCHECK) {
		double bw,time,pwnd_estimate;
		int packets;
		double now = Scheduler::instance().clock();
		switch (algorithm_) {
		case 0:
			packets = last_ack_ - prev_last_ack_;
			time = now - prev_time_;
			if (time == 0) time = 1;
			bw = (double) packets*size_*8/time;
			wbwa_ = weight_ * bw + (1-weight_)*wbwa_;
			prev_last_ack_ = last_ack_;
			prev_time_ = now;
			if (wbwa_ < target_bw_) {
				pwnd_estimate = probability_ * cwnd_;
				pwnd_estimate += time/(t_rtt_*tcp_tick_);
				probability_ = pwnd_estimate/cwnd_;
				if (probability_ > 1) probability_ = 1;
			}
			else {
				pwnd_estimate = probability_ * cwnd_;
				pwnd_estimate -= time/(t_rtt_*tcp_tick_);
				probability_ = pwnd_estimate/cwnd_;
				if (probability_ < 0) probability_ = 0;
			}
			break;
		}
		pwnd_ = probability_ * cwnd_;
		bwnd_ = cwnd_ - pwnd_;
		set_bwcheck_timer();
	}

}

void Sack1PriorityTcpAgent::set_bwcheck_timer()
{
	if (pending_[TCP_TIMER_BWCHECK]) {
		cancel(TCP_TIMER_BWCHECK);
	}
	sched(bwcheck_interval_,TCP_TIMER_BWCHECK);
}
